import datetime
import re
from io import BytesIO
from typing import Any, TypedDict

from dateutil import parser

from shinobi.decorators.return_error_decorator import return_on_error
from shinobi.mixins.parser.base import BaseParser


class StaffImageDictionary(TypedDict):
    image: BytesIO
    mimetype: str


class StaffDictionary(TypedDict):
    mal_id: str
    name: str
    staff_image: StaffImageDictionary
    given_name: str
    family_name: str
    alternate_name: list[str]
    birthday: datetime.datetime
    about: str


class StaffParser(BaseParser):
    def __init__(self, html: str):
        super().__init__()
        self.parser = self.get_parser(html)

    @property
    @return_on_error("")
    def get_staff_url(self) -> str:
        return self.parser.css_first("meta[property='og:url']").attributes["content"]

    @property
    @return_on_error("")
    def get_staff_id(self) -> str:
        return self.regex_helper.get_id_from_url(self.get_staff_url)

    @property
    @return_on_error({})
    def get_staff_image(self) -> StaffImageDictionary:
        url = self.parser.css_first("meta[property='og:image']").attributes["content"]
        if url:
            res = self.client.get(url)
            return {
                "image": BytesIO(res.content),
                "mimetype": url.split(".")[-1],
            }

    @property
    @return_on_error("")
    def get_staff_name(self) -> str:
        return self.parser.css_first("meta[property='og:title']").attributes["content"]

    @property
    @return_on_error("")
    def get_staff_about(self) -> str:
        return self.parser.css_first(
            "div#content table tr td.borderClass .people-informantion-more"
        ).text()

    @property
    @return_on_error("")
    def get_staff_family_name(self) -> str:
        node = self.parser.css_first("#content table tr td.borderClass")
        matches = node.select("span").text_contains("Family name:").matches
        if len(matches) > 1:
            raise ValueError("There are more than one node in family name node")

        # MAL screwed up the HTML here
        html_match = re.search(
            r"Family name:(.*?)(Alternate names|Birthday|Website|Member Favorites|More)",
            matches[0].parent.text(),
        )
        family_name = self.string_helper.cleanse(html_match.group(1))
        # MyAnimeList has it Empty in places
        # raise AttributeError so we can return None
        if not family_name:
            raise AttributeError("Can't find family name using regex")

        return family_name

    @property
    @return_on_error("")
    def get_staff_given_name(self) -> str:
        node = self.parser.select("span").text_contains("Given name:").matches
        if len(node) > 1:
            raise ValueError("There are more than one node in given name node")

        given_name = self.string_helper.cleanse(node[0].next.text())
        if not given_name:
            raise AttributeError("Can't find given name")

        return given_name

    @property
    @return_on_error([])
    def get_staff_alternate_name(self) -> str:
        node = self.parser.css_first("div#content table tr td.borderClass")
        matches = node.select("div span").text_contains("Alternate names:").matches

        if len(matches) > 1:
            raise ValueError("There are more than one node in alternate name node")

        names = matches[0].next.text().split(",")
        alternate_name = [self.string_helper.cleanse(name) for name in names]

        return alternate_name

    @property
    @return_on_error("")
    def get_staff_birthday(self) -> str:
        node = self.parser.css_first("div#content table tr td.borderClass")
        matches = node.select("div span").text_contains("Birthday:").matches
        if len(matches) > 1:
            raise ValueError("There are more than one node in birthday node")

        birthday = parser.parse(self.string_helper.cleanse(matches[0].next.text()))
        return birthday

    def build_dictionary(self) -> dict[str, Any]:
        dictionary: StaffDictionary = {
            "mal_id": self.get_staff_id,
            "name": self.get_staff_name,
            "given_name": self.get_staff_given_name,
            "family_name": self.get_staff_family_name,
            "alternate_name": self.get_staff_alternate_name,
            "birthday": self.get_staff_birthday,
            "about": self.get_staff_about,
            "staff_image": self.get_staff_image,
        }
        return dictionary



import re


class RegexHelper:
    # Getters
    # -------
    def get_id_from_url(self, url: str) -> int:
        pattern = re.compile(r"/(\d+)/")

        _match = re.search(pattern, url)
        if not _match:
            raise Exception("No match found")

        _id_ = _match.group(1)
        if not _id_.isdigit():
            raise Exception("Id is not a digit.")

        return int(_id_)

    def get_content_between_first_brackets(self, text: str) -> str:
        pattern = re.compile(r"\((.*?)\)")
        if content := re.search(pattern, text):
            return content.group(1)
        else:
            raise Exception("No content found")

    def get_first_integer_from_url(self, text: str) -> int:
        pattern = r"\/(\d+)\/"
        _matches = re.search(pattern, text)
        _id = _matches.group(1)
        if not _id.isdigit():
            raise Exception("Id is not a digit.")

        return int(_id)

    # Checks
    # -------
    def check_if_string_contains_integer(self, string: str) -> bool:
        pattern = re.compile(r"\d+")
        return bool(re.search(pattern, string))

    def check_if_string_contains_bracket(self, string: str) -> bool:
        pattern = re.compile(r"\[\d+\]")
        return bool(re.search(pattern, string))

    # Replacements
    # ------------
    def replace_br_with_newline(self, text: str) -> str:
        return re.sub(r"<br\s*\/?>", "\n", text)

    def remove_anime_from_the_end_of_a_string(self, text: str) -> str:
        pattern = r"(.*?)\sAnime"

        return re.sub(pattern, r"\1", text)



from selectolax.parser import HTMLParser
from shinobi.utilities.regex import RegexHelper
from shinobi.utilities.string import StringHelper
from shinobi.utilities.session import session


class BaseParser:
    def __init__(self):
        # Facades
        self.regex_helper = RegexHelper()
        self.string_helper = StringHelper()

        # Client
        self.client = session

    @staticmethod
    def get_parser(html: str) -> HTMLParser:
        return HTMLParser(html)



import string

from selectolax.parser import HTMLParser

from shinobi.decorators.return_error_decorator import return_on_error
from shinobi.utilities.regex import RegexHelper
from shinobi.utilities.session import session
from shinobi.utilities.string import StringHelper


class CharacterBuilder:
    def __init__(self) -> None:
        self.anchors: list[str] = []
        self.visited_urls: set[str] = set()

        # Reusuable clients
        self.client = session

        # Facades
        self.regex_helper = RegexHelper()
        self.string_helper = StringHelper()

    @staticmethod
    def get_parser(html: str) -> HTMLParser:
        return HTMLParser(html)

    @return_on_error("")
    def has_next_page(self, html: str) -> bool:
        parser = self.get_parser(html)
        node = parser.css_first("div.normal_header > div.fl-r > div > span.bgColor1")

        select_node_list = node.text().split(" ")
        for item in select_node_list:
            if self.regex_helper.check_if_string_contains_bracket(item):
                bracketed_element_position = select_node_list.index(item)
                break

        if bracketed_element_position == len(select_node_list) - 1:
            return False

        return True

    def get_all_pages_in_span_tag(self, html: str) -> list[str]:
        parser = self.get_parser(html)
        node = (
            parser.css_first("div.normal_header > div.fl-r > div > span.bgColor1")
            .select("a")
            .matches
        )
        anchors = [anchor.attributes["href"] for anchor in node]
        return [item for item in anchors if item is not None]

    def _build_word_list(self) -> list[str]:
        alphabet_list = list(string.ascii_uppercase)
        return [
            f"https://myanimelist.net/character.php?letter={letter}"
            for letter in alphabet_list
        ]

    def __build_urls(self, url: str) -> None:
        self.visited_urls.add(url)

        res = self.client.get(url)
        html = res.text

        character_nodes = self.get_parser(html).css("a[href*='/character/']")

        if len(character_nodes) == 0:
            print(
                f"""
                Status : {res.status_code}
                URL : {url}
            """
            )

        for character_node in character_nodes:
            character_href = character_node.attributes["href"]
            if (
                character_href
                and character_href not in self.anchors
                and self.regex_helper.check_if_string_contains_integer(character_href)
            ):
                self.anchors.append(
                    self.string_helper.add_myanimelist_if_not_already_there(
                        character_href
                    )
                )

        if self.has_next_page(html):
            all_pages = self.get_all_pages_in_span_tag(html)
            for item in all_pages:
                myanimelist_formated_url = "https://myanimelist.net" + item
                if myanimelist_formated_url not in self.visited_urls:
                    next_url = myanimelist_formated_url
                    break

            self.__build_urls(next_url)

    def __build_ids(self) -> list[int]:
        return [
            self.regex_helper.get_first_integer_from_url(item) for item in self.anchors
        ]

    def build_dictionary(
        self, excluded_ids: list[int] | None = None, sort: bool = False
    ) -> dict[int, str]:
        for url in self._build_word_list():
            self.__build_urls(url)

        dictionary = dict(zip(self.__build_ids(), self.anchors))

        if sort:
            dictionary = dict(sorted(dictionary.items()))

        if excluded_ids:
            dictionary = {
                key: value
                for key, value in dictionary.items()
                if key not in excluded_ids
            }

        return dictionary



# Code Owners : `horidesu`, `baseplate-admin`
# Licensed Under : AGPL-v3
import datetime
from typing import TypedDict

from dateutil import parser
from decorators.return_error_decorator import return_on_error

from shinobi.mixins.parser.base import BaseParser


class ProducerDictionary(TypedDict):
    mal_id: str
    name: str
    name_japanese: str
    established: datetime.datetime
    about: str


class ProducerParser(BaseParser):
    def __init__(self, html: str):
        super().__init__()

        self.parser = self.get_parser(html)

    @property
    @return_on_error("")
    def get_producer_url(self):
        return self.parser.css_first('meta[property="og:url"]').attributes["content"]

    @property
    @return_on_error("")
    def get_producer_id(self) -> str:
        return self.regex_helper.get_id_from_url(self.get_producer_url)

    @property
    @return_on_error("")
    def get_producer_name(self) -> str:
        node = self.parser.css_first("#contentWrapper > div:first-child > h1")
        return self.string_helper.cleanse(node.text())

    @property
    @return_on_error("")
    def get_producer_japanese_name(self) -> str:
        node = self.parser.select("span").text_contains("Japanese:")
        return self.string_helper.cleanse(node.matches[0].next.text())

    @property
    @return_on_error("")
    def get_producter_establish_date(self) -> datetime.datetime:
        node = self.parser.select("span").text_contains("Established:")
        string_date = self.string_helper.cleanse(node.matches[0].next.text())
        actual_date = parser.parse(string_date)
        return actual_date

    @property
    @return_on_error("")
    def get_producer_about(self) -> str:
        return self.parser.css_first(
            "#content > div:nth-of-type(1) div.spaceit_pad > span:not(.dark_text)"
        ).text()

    def build_dictionary(self) -> ProducerDictionary:
        dictionary: ProducerDictionary = {
            "mal_id": self.get_producer_id,
            "name": self.get_producer_name,
            "name_japanese": self.get_producer_japanese_name,
            "established": self.get_producter_establish_date,
            "about": self.get_producer_about,
        }
        return dictionary



# Code Owners : `horidesu`, `baseplate-admin`
# Licensed Under : AGPL-v3

from io import BytesIO
from typing import TypedDict


from shinobi.decorators.return_error_decorator import return_on_error
from shinobi.mixins.parser.base import BaseParser
from selectolax.parser import HTMLParser


class CharacterImageDictionary(TypedDict):
    image: BytesIO
    mimetype: str


class CharacterDictionary(TypedDict):
    mal_id: str
    name: str
    name_kanji: str
    character_image: CharacterImageDictionary
    about: str


class CharacterParser(BaseParser):
    def __init__(self, html: str):
        super().__init__()

        self.parser = self.get_parser(html)

    @property
    @return_on_error("")
    def get_character_url(self) -> str:
        if character_url := self.parser.css_first("meta[property='og:url']").attributes[
            "content"
        ]:
            return character_url
        else:
            raise AttributeError("`get_character_url` element not found")

    @property
    @return_on_error("")
    def get_character_id(self) -> str:
        if character_id := self.regex_helper.get_id_from_url(self.get_character_url):
            return character_id
        else:
            raise AttributeError("`get_character_id` element not found")

    @property
    @return_on_error("")
    def get_character_name(self) -> str:
        return self.parser.css_first("meta[property='og:title']").attributes["content"]

    @property
    @return_on_error("")
    def get_character_name_kanji(self) -> str:
        return self.regex_helper.get_content_between_first_brackets(
            self.parser.css_first("h2.normal_header span small").text()
        )

    @property
    @return_on_error("")
    def get_about(self) -> str:
        html = self.parser.css_first("#content table tbody tr > td:nth-of-type(2)")
        tags = ["div", "br", "table", "h2", "script"]
        html.strip_tags(tags)

        sentences = html.text().split("\n")
        return "\n\n".join(sentences).strip()

    @property
    @return_on_error("")
    def get_character_image(self) -> str:
        url = self.parser.css_first("meta[property='og:image']").attributes["content"]
        if url:
            res = self.client.get(url)
            return {
                "image": BytesIO(res.content),
                "mimetype": url.split(".")[-1],
            }

    def build_dictionary(self) -> CharacterDictionary:
        dictionary: CharacterDictionary = {
            "mal_id": self.get_character_id,
            "name": self.get_character_name,
            "name_kanji": self.get_character_name_kanji,
            "character_image": self.get_character_image,
            "about": self.get_about,
        }
        return dictionary



import datetime
from functools import lru_cache
from typing import TypedDict

from dateutil import parser
from selectolax.parser import Node

from shinobi.decorators.return_error_decorator import return_on_error
from shinobi.mixins.parser.base import BaseParser


class AnimeDictionary(TypedDict):
    mal_id: int
    name: str
    name_japanese: str
    name_synonyms: list[str]
    source: str
    aired_from: datetime.datetime
    aired_to: datetime.datetime
    synopsis: str
    background: str
    rating: str
    genres: list[int]
    themes: list[int]
    studios: list[int]
    producers: list[int]
    demographics: list[int]
    recommendations: list[int]
    openings: list[int]
    endings: list[int]


class AnimeParser(BaseParser):
    def __init__(self, html: str) -> None:
        super().__init__()

        self.parser = self.get_parser(html)

    @property
    @return_on_error("")
    @lru_cache(maxsize=None)
    def __get_aired_text(self):
        # aired text contains in this format
        # 'aired_from to aired_to'
        node = self.parser.select("span").text_contains("Aired:").matches
        if len(node) > 1:
            raise ValueError("There are multiple aired node")

        return self.string_helper.cleanse(node[0].next.text())

    @property
    @return_on_error("")
    def get_anime_url(self) -> str:
        if anime_url := self.parser.css_first("meta[property='og:url']").attributes[
            "content"
        ]:
            return anime_url
        else:
            raise AttributeError("`get_anime_url` element not found")

    @property
    @return_on_error("")
    def get_anime_id(self) -> str:
        if anime_id := self.regex_helper.get_id_from_url(self.get_anime_url):
            return str(anime_id)
        else:
            raise AttributeError("`get_anime_id` element not found")

    @property
    @return_on_error("")
    def get_anime_name(self) -> str:
        if anime_name := self.parser.css_first("meta[property='og:title']").attributes[
            "content"
        ]:
            return anime_name
        else:
            raise AttributeError("`get_anime_name` element not found")

    @property
    @return_on_error("")
    def get_anime_name_japanese(self) -> str:
        node = self.parser.select("span").text_contains("Japanese:").matches
        if len(node) > 1:
            raise ValueError("There are more than one node in name japanese node")

        name_japanese = self.string_helper.cleanse(node[0].next.text())
        return name_japanese

    @property
    @return_on_error([])
    def get_anime_name_synonyms(self) -> list[str]:
        node = self.parser.select("h2").text_contains("Alternative Titles").matches[0]
        alternate_names = []

        next_node: Node

        while True:
            if node.next.tag == "h2":
                break
            elif node.next.tag == "div":
                if node.next:
                    next_node = node.next

            try:
                alternate_name = self.string_helper.cleanse(
                    next_node.css_first("span").next.text()
                )
                alternate_names.append(alternate_name)

                next_node.decompose(recursive=True)

            # There are no nodes
            except AttributeError:
                break

        return alternate_names

    @property
    @return_on_error("")
    def get_source(self) -> str:
        node = self.parser.select("span").text_contains("Source:").matches
        if len(node) > 1:
            raise ValueError("There are multiple source node")

        source = self.string_helper.cleanse(node[0].next.text())
        return source

    @property
    def get_aired_from(self) -> datetime.datetime:
        aired_text = self.__get_aired_text
        splitted_text = aired_text.split("to")
        aired_from = parser.parse(self.string_helper.cleanse(splitted_text[0]))
        return aired_from

    @property
    def get_aired_to(self) -> datetime.datetime:
        aired_text = self.__get_aired_text
        splitted_text = aired_text.split("to")
        aired_to = parser.parse(self.string_helper.cleanse(splitted_text[1]))
        return aired_to

    @property
    @return_on_error("")
    def get_synopsis(self) -> str:
        node = self.parser.css_first("p[itemprop='description']")

        synopsis = self.string_helper.cleanse(node.text())
        return (
            ""
            if "No synopsis information has been added to this title." in synopsis
            else synopsis
        )

    @property
    @return_on_error("")
    def get_background(self) -> str:
        node = self.parser.select("h2").text_contains("Background").matches
        if len(node) > 1:
            raise ValueError("There are multiple Background node")

        parent_node = node[0].parent.parent
        parent_node.strip_tags(["p", "div"])

        background = self.string_helper.cleanse(parent_node.text())
        return background

    @property
    @return_on_error("")
    def get_rating(self) -> str:
        node = self.parser.select("span").text_contains("Rating:").matches
        if len(node) > 1:
            raise ValueError("There are multiple Rating node")

        rating = self.string_helper.cleanse(node[0].next.text())
        return rating

    @property
    @return_on_error([])
    def get_genres(self) -> list[int]:
        node = self.parser.select("span").text_contains("Genres:").matches
        if len(node) > 1:
            raise ValueError("There are multiple Genre node")

        genre_parent_nodes = node[0].parent
        # Remove all span tags
        anchor_tags = genre_parent_nodes.css("a[href*='/anime/']")
        return sorted(
            {
                self.regex_helper.get_first_integer_from_url(anchor.attributes["href"])
                for anchor in anchor_tags
                if anchor.attributes["href"]
            }
        )

    @property
    @return_on_error([])
    def get_themes(self) -> list[int]:
        node = self.parser.select("span").text_contains("Themes:").matches
        if len(node) > 1:
            raise ValueError("There are multiple Genre node")

        theme_parent_nodes = node[0].parent
        anchor_tags = theme_parent_nodes.css("a[href*='/anime/']")
        return sorted(
            {
                self.regex_helper.get_first_integer_from_url(anchor.attributes["href"])
                for anchor in anchor_tags
                if anchor.attributes["href"]
            }
        )

    @property
    @return_on_error([])
    def get_studios(self) -> list[int]:
        node = self.parser.select("span").text_contains("Studios:").matches
        if len(node) > 1:
            raise ValueError("There are multiple Studio node")

        anchor_tags = node[0].parent.css("a[href*='/producer/']")
        return sorted(
            {
                self.regex_helper.get_first_integer_from_url(anchor.attributes["href"])
                for anchor in anchor_tags
                if anchor.attributes["href"]
            }
        )

    @property
    @return_on_error([])
    def get_producers(self) -> list[int]:
        node = self.parser.select("span").text_contains("Producers:").matches
        if len(node) > 1:
            raise ValueError("There are multiple Producer node")

        anchor_tags = node[0].parent.css("a[href*='/producer/']")
        return sorted(
            {
                self.regex_helper.get_first_integer_from_url(anchor.attributes["href"])
                for anchor in anchor_tags
                if anchor.attributes["href"]
            }
        )

    @property
    @return_on_error([])
    def get_demographics(self) -> list[int]:
        node = self.parser.select("span").text_contains("Demographic:").matches
        if len(node) > 1:
            raise ValueError("There are multiple Demographic node")

        anchor_tags = node[0].parent.css("a[href*='/genre/']")
        return sorted(
            {
                self.regex_helper.get_first_integer_from_url(anchor.attributes["href"])
                for anchor in anchor_tags
                if anchor.attributes["href"]
            }
        )

    @property
    @return_on_error({})
    def get_openings(self) -> dict[int, str]:
        node = self.parser.select("h2").text_contains("Opening Theme").matches

        opening_table_tag = node[0].parent.next.next
        # stupid myanimelist shows popup as table
        opening_table_tag.strip_tags(["div#js-oped-popup"])
        openings: dict[int, str] = {}

        table_rows = opening_table_tag.css("tr")
        for row in table_rows:
            colon_spilitted_text = row.text().split(":")
            # There are no colons in text
            # "Hikaru nara"
            if len(colon_spilitted_text) == 1:
                openings[1] = self.string_helper.cleanse(colon_spilitted_text[0])

            # "1: Hikaru nara"
            else:
                openings[int(colon_spilitted_text[0])] = self.string_helper.cleanse(
                    colon_spilitted_text[1]
                )

        return openings

    @property
    @return_on_error({})
    def get_endings(self) -> dict[int, str]:
        node = self.parser.select("h2").text_contains("Ending Theme").matches

        endings_table_tag = node[0].parent.next.next
        # stupid myanimelist shows popup as table
        endings_table_tag.strip_tags(["div#js-oped-popup"])
        endings: dict[int, str] = {}

        table_rows = endings_table_tag.css("tr")
        for row in table_rows:
            colon_spilitted_text = row.text().split(":")
            # There are no colons in text
            # "Hikaru nara"
            if len(colon_spilitted_text) == 1:
                endings[1] = self.string_helper.cleanse(colon_spilitted_text[0])

            # "1: Hikaru nara"
            else:
                endings[int(colon_spilitted_text[0])] = self.string_helper.cleanse(
                    colon_spilitted_text[1]
                )

        return endings

    def build_dictionary(self) -> AnimeDictionary:
        dictionary: AnimeDictionary = {
            "mal_id": self.get_anime_id,
            "name": self.get_anime_name,
            "name_japanese": self.get_anime_name_japanese,
            "name_synonyms": self.get_anime_name_synonyms,
            "source": self.get_source,
            "aired_from": self.get_aired_from,
            "aired_to": self.get_aired_to,
            "synopsis": self.get_synopsis,
            "background": self.get_background,
            "rating": self.get_rating,
            "genres": self.get_genres,
            "themes": self.get_themes,
            "studios": self.get_studios,
            "producers": self.get_producers,
            "demographics": self.get_demographics,
            "recommendations": [],  # self
            "openings": self.get_openings,
            "endings": self.get_endings,
        }
        return dictionary



import os
from typing import Any

from redis import ConnectionPool
from requests import Session
from requests.adapters import HTTPAdapter
from requests.structures import CaseInsensitiveDict
from requests.utils import DEFAULT_ACCEPT_ENCODING
from requests_cache import RedisCache
from requests_ratelimiter import LimiterMixin, RedisBucket
from urllib3.util import Retry


class CachedLimiterSession(LimiterMixin, Session):
    """
    Session class with caching and rate-limiting behavior.
        Accepts arguments for both LimiterSession and CachedSession.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.headers = self.default_headers()

    def default_headers(self) -> CaseInsensitiveDict[str | bytes]:
        """
        :rtype: requests.structures.CaseInsensitiveDict
        """
        return CaseInsensitiveDict(
            {
                "User-Agent": "CoreProject",
                "Accept-Encoding": DEFAULT_ACCEPT_ENCODING,
                "Accept": "*/*",
                "Connection": "keep-alive",
            }
        )


CACHE_NAME = "seeder"
RETRY_STATUSES = [403]

retry_strategy = Retry(
    total=15,
    backoff_factor=2,
    status_forcelist=RETRY_STATUSES,
)
adapter = HTTPAdapter(max_retries=retry_strategy)
redis_pool = ConnectionPool.from_url(os.environ.get("REDIS_URL", "redis://localhost:6379"))

session = CachedLimiterSession(
    bucket_class=RedisBucket,
    backend=RedisCache(),
    # Undocumented ( pyrate-limiter )
    bucket_kwargs={
        "redis_pool": redis_pool,
        "bucket_name": CACHE_NAME,
        "expire_time": 30,
    },
    # https://docs.api.jikan.moe/#section/Information/Rate-Limiting
    per_minute=100,
    # per_second=1,
    per_host=True,
    # https://requests-cache.readthedocs.io/en/stable/user_guide/expiration.html
    expire_after=360,
)
session.mount("http://", adapter)
session.mount("https://", adapter)



from selectolax.parser import HTMLParser

from shinobi.utilities.regex import RegexHelper
from shinobi.utilities.session import session


class AnimeGenreBuilder:
    def __init__(self) -> None:
        # Reusuable Clients
        self.client = session

        # Facades
        self.regex_helper = RegexHelper()

    @staticmethod
    def get_parser(html: str) -> HTMLParser:
        return HTMLParser(html)

    def __build_urls(self, ids: list[int]) -> list[str]:
        return [f"https://myanimelist.net/anime/genre/{mal_id}" for mal_id in ids]

    def __build_ids(self, html: str) -> list[int]:
        parser = self.get_parser(html)
        genre_nodes = parser.css('a[href*="genre"]')
        link_ids = [
            self.regex_helper.get_id_from_url(node.attributes["href"])
            for node in genre_nodes
            if node.attributes["href"]
        ]

        return sorted(link_ids)

    def build_dictionary(self) -> dict[int, str]:
        res = self.client.get("https://myanimelist.net/anime.php")
        html = res.text

        ids = self.__build_ids(html)
        urls = self.__build_urls(ids)

        dictionary = dict(zip(ids, urls))

        return dictionary



import re
from selectolax.parser import HTMLParser
from typing import Set
from shinobi.utilities.regex import RegexHelper
from shinobi.utilities.session import session

import xml.etree.ElementTree as ET


class StaffBuilder:
    def __init__(self) -> None:
        # Reusuable clients
        self.client = session

        # Facades
        self.regex_helper = RegexHelper()

        # Urls
        self.urls_to_visit = self.__build_people()

    @staticmethod
    def get_parser(html: str) -> HTMLParser:
        return HTMLParser(html)

    def __build_people(self) -> Set[int]:
        url = "https://myanimelist.net/sitemap/index.xml"
        res = self.client.get(url)
        tree = ET.fromstring(res.content)

        people_urls = []
        pattern = re.compile(r"people")

        people_urls = [
            element.text
            for sitemap in tree
            for element in sitemap
            if "loc" in element.tag and pattern.search(element.text)
        ]

        return set(people_urls)

    def build_dictionary(
        self, excluded_ids: list[int] | None = None, sort: bool = False
    ) -> dict[int, str]:
        dictionary = {}

        for url in self.urls_to_visit:
            res = self.client.get(url)
            tree = ET.fromstring(res.content)

            for entry in tree:
                for element in entry:
                    if "loc" in element.tag:
                        content = element.text
                        dictionary[self.regex_helper.get_id_from_url(content)] = content

        if sort:
            dictionary = dict(sorted(dictionary.items()))

        return dictionary



import html
import re


class StringHelper:
    def cleanse(self, string: str) -> str:
        # Convert any HTML line breaks to newlines
        string = re.sub(r"<br\s*/?>", "\n", string)

        # Convert non-breaking spaces to spaces
        string = string.replace("\xa0", " ")

        # Strip any remaining HTML tags
        string = html.unescape(string)  # Decode HTML entities
        string = re.sub(r"<[^>]*>", "", string)

        # Remove newlines at the end
        string = string.rstrip("\n")

        # Trim whitespace
        string = string.strip()

        return string

    @staticmethod
    def add_myanimelist_if_not_already_there(url: str) -> str:
        if "myanimelist.net" not in url:
            return "https://myanimelist.net" + url
        else:
            return url



import functools
from typing import Any, Callable

# These decorators catch :
#   AttributeError : In case `selectolax` fails to find the dom node
#   IndexError : In case `selectolax` finds empty dom node


def return_on_error[T](return_type: T) -> Callable[[Callable], T]:
    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> T:
            try:
                return func(*args, **kwargs)
            except (AttributeError, IndexError):
                return return_type

        return wrapper

    return decorator



from typing import TypedDict


from shinobi.decorators.return_error_decorator import return_on_error
from shinobi.mixins.parser.base import BaseParser


class GenreDictionary(TypedDict):
    mal_id: int
    name: str
    description: str
    type: str


class AnimeGenreParser(BaseParser):
    def __init__(self, html: str):
        super().__init__()

        self.parser = self.get_parser(html)

    @property
    @return_on_error("")
    def get_url(self) -> str:
        return self.parser.css_first('meta[property="og:url"]').attributes["content"]

    @property
    @return_on_error("")
    def get_mal_id(self) -> int:
        return self.regex_helper.get_id_from_url(self.get_url)

    @property
    @return_on_error("")
    def get_name(self) -> str:
        html = self.parser.css_first("span.di-ib.mt4")
        # Remove span nodes
        html.strip_tags(["span.fw-n"])
        actual_text = self.regex_helper.remove_anime_from_the_end_of_a_string(html.text())
        return self.string_helper.cleanse(actual_text)

    @property
    @return_on_error("")
    def get_description(self) -> str:
        text = self.parser.css_first("p.genre-description").text()
        return self.regex_helper.replace_br_with_newline(text)

    @property
    @return_on_error("")
    def get_type(self) -> str:
        text = self.parser.css("div.di-ib")
        # The second node actually shows the item
        actual_node = text[1]
        return self.string_helper.cleanse(actual_node.text())

    def build_dictionary(self) -> GenreDictionary:
        dictionary: GenreDictionary = {
            "mal_id": self.get_mal_id,
            "name": self.get_name,
            "description": self.get_description,
            "type": self.get_type,
        }
        return dictionary



from selectolax.parser import HTMLParser

from shinobi.utilities.regex import RegexHelper
from shinobi.utilities.session import session
from shinobi.utilities.string import StringHelper


class AnimeThemeBuilder:
    def __init__(self) -> None:
        self.anchors: list[str] = []

        # Reusuable clients
        self.client = session

        # Facades
        self.regex_helper = RegexHelper()
        self.string_helper = StringHelper()

    @staticmethod
    def get_parser(html: str) -> HTMLParser:
        return HTMLParser(html)

    def __build_ids(self) -> list[int]:
        return [self.regex_helper.get_first_integer_from_url(item) for item in self.anchors]

    def __build_urls(self, html: str) -> list[str]:
        parser = self.get_parser(html)
        theme_parent_node = (
            parser.select("div.normal_header").text_contains("Themes").matches[0].next.next
        )
        theme_anchor_nodes = theme_parent_node.css('a[href*="genre"]')

        self.anchors = [
            self.string_helper.add_myanimelist_if_not_already_there(
                anchor.attributes["href"]
            )
            for anchor in theme_anchor_nodes
            if anchor.attributes["href"]
        ]
        return self.anchors

    def build_dictionary(self) -> dict[int, str]:
        res = self.client.get("https://myanimelist.net/anime.php")
        html = res.text

        urls = self.__build_urls(html)
        ids = self.__build_ids()

        dictionary = dict(zip(ids, urls))

        return dictionary



from shinobi.decorators.return_error_decorator import return_on_error

from shinobi.mixins.parser.base import BaseParser


class AnimeCharacterAndStaffListDictionary:
    characters: list[int]
    staffs: list[int]


class AnimeCharacterAndStaffListParser(BaseParser):
    def __init__(self, html: str) -> None:
        super().__init__()
        self.parser = self.get_parser(html)

    @property
    @return_on_error([])
    def get_characters(self) -> list[int]:
        anchor_tags = self.parser.css("a[href*='/character/']")
        return sorted(
            {
                self.regex_helper.get_first_integer_from_url(anchor.attributes["href"])
                for anchor in anchor_tags
            }
        )

    @property
    @return_on_error([])
    def get_staffs(self) -> list[int]:
        anchor_tags = self.parser.css("a[href*='/people/']")
        return sorted(
            {
                self.regex_helper.get_first_integer_from_url(anchor.attributes["href"])
                for anchor in anchor_tags
            }
        )

    def build_dictionary(self) -> AnimeCharacterAndStaffListDictionary:
        dictionary = {
            "characters": self.get_characters,
            "staffs": self.get_staffs,
        }
        return dictionary
